from abc import ABC, abstractmethod
from pathlib import Path


class Operation(ABC):
    """Base class for all operations"""

    @abstractmethod
    def description(self) -> str:
        """Human-readable description of the operation"""
        pass

    @abstractmethod
    def can_handle(self, path: Path) -> bool:
        """Check if this operation can handle the given file"""
        pass

    @abstractmethod
    def apply(self, content: str) -> str:
        """Apply the operation to the given content and return the modified content"""
        pass



import logging
from pathlib import Path

from django_new.transformer.operations import Operation

logger = logging.getLogger(__name__)


class Transformation:
    """Base class for transformations"""

    def __init__(self, root_path: Path):
        self.root_path = Path(root_path)
        self._changes = []

    def forwards(self):
        """Apply the migration"""

        raise NotImplementedError

    def backwards(self):
        """Reverse the migration"""

        raise NotImplementedError

    def assert_path_is_valid(self, path: Path):
        if not path.exists():
            raise FileNotFoundError(f"File not found: {path}")

        # Check if the path is within the root directory
        try:
            path.relative_to(self.root_path.resolve())
        except ValueError as e:
            raise ValueError(f"Path '{path}' is not within the project root '{self.root_path}'") from e

    def modify_file(self, path: str | Path, operation: Operation):
        """Apply an operation to a file"""

        if isinstance(path, str):
            path = Path(path)

        path = (self.root_path / path).resolve()
        self.assert_path_is_valid(path)

        if not operation.can_handle(path):
            raise ValueError(f"Operation {type(operation).__name__} cannot handle file {path}")

        # Read current content
        content = path.read_text()

        # Store original for rollback
        self._changes.append((path, content))

        # Apply operation
        new_content = operation.apply(content)

        # Write new content to file
        path.write_text(new_content)

    def rollback_changes(self):
        """Rollback all changes made during this session"""

        for path, original_content in reversed(self._changes):
            path.write_text(original_content)

        self._changes.clear()


class Runner:
    """Runs transformations."""

    def __init__(self, path: Path, dry_run: bool = False):  # noqa: FBT001, FBT002
        self.path = path
        self.dry_run = dry_run

        # Track operations in dry-run mode
        self._operations = []

    def install(self, *transformations: list[Transformation]) -> bool:
        """Run forwards transformations."""

        for transformation in transformations:
            if self.dry_run:
                # Intercept operations for dry run
                original_modify = transformation.modify_file

                def track_operation(filepath, operation):
                    self._operations.append((filepath, operation))

                transformation.modify_file = track_operation

                try:
                    transformation.forwards()

                    return self._operations
                finally:
                    transformation.modify_file = original_modify
            else:
                try:
                    transformation.forwards()
                except Exception as e:
                    logger.exception(e)
                    transformation.rollback_changes()

                    raise

        return True

    def uninstall(self, *transformations: list[Transformation]) -> bool:
        """Run backwards transformations."""

        for transformation in transformations:
            if self.dry_run:
                original_modify = transformation.modify_file

                def track_operation(filepath, operation):
                    self._operations.append((filepath, operation))

                transformation.modify_file = track_operation

                try:
                    transformation.backwards()

                    return self._operations
                finally:
                    transformation.modify_file = original_modify
            else:
                try:
                    transformation.backwards()

                    return True
                except Exception as e:
                    logger.exception(e)
                    transformation.rollback_changes()

                    raise

        return True



import logging
import os
from io import StringIO

from rich.console import Console

try:
    from django.core.management import call_command as django_call_command
except ImportError as exc:
    # This should never happen because `Django` is a dependency of `django-new`
    raise ImportError("Couldn't import Django. Are you sure it's installed?") from exc


logger = logging.getLogger(__name__)

console = Console()


def stdout(message: str):
    console.print(message)


def stderr(message: str):
    console.print(message, style="red")


def call_command(*args) -> tuple[str, str]:
    """Call a Django management command and capture its output.

    Args:
        *args: Command name and arguments to pass to the command.

    Returns:
        A tuple of (stdout, stderr).
    """

    # Redirect stdout and stderr to capture the output
    out = StringIO()
    err = StringIO()

    logger.debug(f"Call command with args: {args}")

    django_call_command(*args, stdout=out, stderr=err)

    stdout_str = out.getvalue()
    stderr_str = err.getvalue()

    return (stdout_str, stderr_str)


def is_running_under_any_uv():
    """
    Check if the current environment is running under any version of uv.

    Returns:
        bool: True if running under uv, False otherwise.
    """

    uv_env_vars = (
        "UV_PROJECT_ENVIRONMENT",
        "UV_INTERNAL__PARENT_INTERPRETER",
        "UV_PYTHON",
    )

    return any(os.getenv(var) for var in uv_env_vars)



import logging
from dataclasses import dataclass, field
from importlib import resources
from pathlib import Path
from typing import Any

from django.template import Context, Engine

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class TemplateFile:
    """A template file to be rendered.

    The template file extension should end with '-tpl'.
    """

    path: Path
    context: dict[str, Any] = field(default_factory=dict)


def create_file(
    template_file: TemplateFile,
    resource_name: str = "django_new",
    resource_path: str = "templates",
):
    """Create file based on a DTL template in a specified resource."""

    logger.debug(f"Create file, {template_file.path}, if it doesn't exist")

    if template_file.path.exists():
        logger.debug(f"Do not create template file, {template_file.path}, because it already exists")
    else:
        engine = Engine(debug=False, autoescape=False)

        template_name = template_file.path.name + "-tpl"
        logger.debug(f"Template name: {template_name}")

        template_path = resources.files(resource_name) / resource_path / template_name
        template_content = template_path.read_text()
        logger.debug("Read template content")

        template_content = "{% autoescape off %}" + template_content + "{% endautoescape %}"
        logger.debug("Wrap template content in autoescape off")

        template = engine.from_string(template_content)
        rendered_content = template.render(Context(template_file.context or {}))
        logger.debug(f"Render template content with context {template_file.context}")

        template_file.path.write_text(rendered_content)
        logger.debug(f"Created template file, {template_file.path}")



import logging
from pathlib import Path

from django_new.parser import get_class_name
from django_new.templater.django_template import TemplateFile, create_file
from django_new.transformer import Transformation
from django_new.transformer.operations.python import AppendToList
from django_new.utils import call_command, stdout

logger = logging.getLogger(__name__)


CLASSIC_CONFIGURATION_PATH_NAME = "config"


class AppCreator:
    def __init__(self, app_name: str | None, folder: Path):
        self.app_name = app_name

        if self.app_name is None:
            self.app_name = self.default_app_name

        if not self.app_name:
            raise ValueError("App name is unknown")

        self.folder = folder

    def create(self) -> None:
        """Create a new Django app."""

        logger.debug(f"Start creating app, {self.app_name}")

        logger.debug(f"Create app directory, {self.folder / self.app_name}, if it doesn't exist")
        (self.folder / self.app_name).mkdir(parents=True, exist_ok=True)

        call_command("startapp", self.app_name, self.folder / self.app_name)

        # Remove tests.py in lieu of a root directory named tests
        (self.folder / self.app_name / "tests.py").unlink(missing_ok=True)

        # Create tests directory with __init__.py
        tests_dir = self.folder / "tests" / self.app_name
        tests_dir.mkdir(parents=True, exist_ok=True)
        (tests_dir / "__init__.py").touch(exist_ok=True)
        logger.debug(f"Created tests directory at {tests_dir}")

        settings_path = self.folder / "settings.py"

        # If settings cannot be found, look in "config" folder
        if not settings_path.exists():
            settings_path = self.folder / CLASSIC_CONFIGURATION_PATH_NAME / "settings.py"

        if settings_path.exists():
            apps_path = self.folder / self.app_name / "apps.py"

            self.add_app_to_installed_apps(name=self.app_name, apps_path=apps_path, settings_path=settings_path)

    def add_app_to_installed_apps(self, name: str, apps_path: Path, settings_path: Path):
        logger.debug(f"Add {name} to INSTALLED_APPS")

        if apps_path.exists():
            app_config_name = get_class_name(path=apps_path, base_class_name="AppConfig")

            if app_config_name:
                fully_qualified_app_config_name = f"{name}.apps.{app_config_name}"

                transformer = Transformation(root_path=self.folder)
                operation = AppendToList(name="INSTALLED_APPS", value=f'"{fully_qualified_app_config_name}"')
                transformer.modify_file(path=settings_path, operation=operation)

                stdout(f"✅ [blue]{fully_qualified_app_config_name}[/blue] added to [blue]INSTALLED_APPS[/blue]")
            else:
                logger.error("app_config_name could not be determined")
        else:
            logger.error("apps.py doesn't exist")


class ApiAppCreator(AppCreator):
    default_app_name = "api"

    def create(self) -> None:
        super().create()

        # Add urls.py
        urls_template_file = TemplateFile(self.folder / self.app_name / "urls.py", {"app_name": self.app_name})
        create_file(template_file=urls_template_file, resource_path="templates/app_template")


class WebAppCreator(AppCreator):
    default_app_name = "web"

    def create(self) -> None:
        super().create()

        # Create project-level folder for static files
        if not (self.folder / "static").exists():
            (self.folder / "static/css").mkdir(parents=True, exist_ok=True)
            (self.folder / "static/js").mkdir(parents=True, exist_ok=True)
            (self.folder / "static/img").mkdir(parents=True, exist_ok=True)

        # Create urls.py
        urls_template_file = TemplateFile(self.folder / self.app_name / "urls.py", {"app_name": self.app_name})
        create_file(template_file=urls_template_file, resource_path="templates/app_template")

        # Create folder for templates
        (self.folder / self.app_name / "templates" / self.app_name).mkdir(parents=True, exist_ok=True)
        urls_template_file = TemplateFile(
            self.folder / self.app_name / "templates" / self.app_name / "index.html",
            {"app_name": self.app_name},
        )
        create_file(template_file=urls_template_file, resource_path="templates/app_template")

        # Create folder for templatetags
        (self.folder / self.app_name / "templatetags").mkdir(parents=True, exist_ok=True)
        (self.folder / self.app_name / "templatetags" / "__init__.py").touch(exist_ok=True)


class WorkerAppCreator(AppCreator):
    default_app_name = "worker"

    def create(self) -> None:
        super().create()

        # Create tasks.py
        create_file(template_file=TemplateFile(path=self.folder / self.app_name / "tasks.py"))

        # Remove default views.py
        (self.folder / self.app_name / "views.py").unlink(missing_ok=True)



from pathlib import Path
from typing import Any

import tomlkit

from django_new.transformer.operations import Operation


class TomlOperation(Operation):
    """Base class for all TOML operations"""

    def can_handle(self, path: Path) -> bool:
        return path.suffix == ".toml"


class AddKeyValue(TomlOperation):
    """Add a key-value pair to a TOML file"""

    def __init__(self, name: str, key: str, value: Any):
        self.name = name
        self.key = key
        self.value = value

    def description(self) -> str:
        return f"Add {self.key} = {self.value!r} to [{self.name}]"

    def apply(self, content: str) -> str:
        """Add a key-value pair to TOML"""

        doc = tomlkit.parse(content)

        # Navigate to the table
        keys = self.name.split(".")
        current = doc

        for key in keys:
            if key not in current:
                current[key] = {}

            current = current[key]

        # Add the value
        current[self.key] = self.value

        return tomlkit.dumps(doc)


class RemoveKey(TomlOperation):
    """Remove a key from a TOML file"""

    def __init__(self, table_path: str, key: str):
        self.table_path = table_path
        self.key = key

    def description(self) -> str:
        return f"Remove {self.key} from [{self.table_path}]"

    def apply(self, content: str) -> str:
        """Remove a key from TOML"""

        doc = tomlkit.parse(content)

        # Navigate to the table
        keys = self.table_path.split(".")
        current = doc

        for key in keys:
            if key not in current:
                raise ValueError(f"Table path '{self.table_path}' not found")

            current = current[key]

        # Remove the key
        if self.key not in current:
            raise ValueError(f"Key '{self.key}' not found in '{self.table_path}'")

        del current[self.key]

        return tomlkit.dumps(doc)


class AppendToList(TomlOperation):
    """Append a value to a list in a TOML file.

    The key can be a simple key (e.g., "dependencies") or use dot notation
    to specify nested tables (e.g., "project.dependencies").
    """

    def __init__(self, name: str, value: Any):
        self.name = name
        self.value = value
        self._name_parts = name.split(".")
        self._list_key = self._name_parts[-1]
        self._table_name = ".".join(self._name_parts[:-1]) if len(self._name_parts) > 1 else ""

    def description(self) -> str:
        return f"Append {self.value!r} to {self.name}"

    def apply(self, content: str) -> str:
        doc = tomlkit.parse(content)
        current = doc

        # Navigate to the parent table if needed
        if self._table_name:
            for key in self._table_name.split("."):
                if key not in current:
                    current[key] = tomlkit.table()
                current = current[key]

        # Handle the list operation
        if self._list_key not in current:
            current[self._list_key] = tomlkit.array([self.value])
        else:
            if not isinstance(current[self._list_key], list | tomlkit.items.Array):
                raise ValueError(f"Cannot append to '{self._list_key}': target is not a list")
            current[self._list_key].append(self.value)

        return tomlkit.dumps(doc)



from pathlib import Path
from typing import Any

import libcst as cst

from django_new.transformer.operations import Operation


class PythonOperation(Operation):
    """Base class for all Python operations"""

    def can_handle(self, path: Path) -> bool:
        return path.suffix.lower() == ".py"


class AppendToList(PythonOperation):
    """Append a value to a Python list, supporting nested class traversal"""

    class AddToListTransformer(cst.CSTTransformer):
        """CST transformer to add items to a list, supporting nested class traversal"""

        def __init__(self, name: str, value: str, position: int | None, after: str | None):
            self.name = name.split(".")
            self.value = value
            self.position = position
            self.after = after
            self.found = False
            self.current_name = []

        def visit_ClassDef(self, node: cst.ClassDef) -> bool | None:  # noqa: N802
            self.current_name.append(node.name.value)

            return True

        def leave_ClassDef(self, _: cst.ClassDef, updated_node: cst.CSTNode) -> cst.CSTNode:  # noqa: N802
            if self.current_name:
                self.current_name.pop()

            return updated_node

        def leave_Assign(self, _: cst.Assign, updated_node: cst.Assign) -> cst.CSTNode:  # noqa: N802
            # Skip if we're not at the right path depth
            # For nested attributes, the path length should match the current name parts
            # e.g., for 'Settings.INSTALLED_APPS', we want to match when current_name is ['Settings']
            if len(self.current_name) != len(self.name) - 1:
                return updated_node

            # Check if this is an assignment to our target list
            target = updated_node.targets[0].target

            # Handle direct name access (e.g., INSTALLED_APPS)
            if isinstance(target, cst.Name):
                target_name = target.value

                # If we're in a class context, we need to check if the class name matches
                if self.current_name and self.current_name[0] == self.name[0]:
                    if target_name == self.name[-1] and isinstance(updated_node.value, cst.List):
                        self.found = True

                        return self._add_to_list_node(updated_node)

                # If not in a class context, just match the name
                elif len(self.name) == 1 and target_name == self.name[0] and isinstance(updated_node.value, cst.List):
                    self.found = True

                    return self._add_to_list_node(updated_node)

            # Handle nested attributes (e.g., Settings.INSTALLED_APPS)
            elif isinstance(target, cst.Attribute):
                # Check if the attribute path matches our target
                attr_parts = []
                node = target

                while isinstance(node, cst.Attribute):
                    attr_parts.append(node.attr.value)
                    node = node.value

                if isinstance(node, cst.Name):
                    attr_parts.append(node.value)
                    attr_parts.reverse()  # Reverse to get the correct order

                    # Check if the full path matches our target
                    if ".".join(attr_parts) == ".".join(self.name) and isinstance(updated_node.value, cst.List):
                        self.found = True
                        return self._add_to_list_node(updated_node)

            return updated_node

        def _add_to_list_node(self, node):
            """Add the value to the list node"""

            # Parse the value as a CST element
            value_node = cst.parse_expression(self.value)
            new_element = cst.Element(value=value_node)

            # Get existing elements
            elements = list(node.value.elements)

            # Insert at position
            if self.after:
                insert_pos = 0
                for i, element in enumerate(elements):
                    # Get the code for the element value to compare
                    element_code = cst.Module([]).code_for_node(element.value).strip()
                    if element_code == self.after:
                        insert_pos = i + 1
                        break
                elements.insert(insert_pos, new_element)
            elif self.position is None:
                elements.append(new_element)
            elif self.position < 0:
                # Negative indexing
                insert_pos = len(elements) + self.position + 1
                elements.insert(max(0, insert_pos), new_element)
            else:
                elements.insert(min(self.position, len(elements)), new_element)

            # Return updated node
            return node.with_changes(value=node.value.with_changes(elements=elements))

    def __init__(self, name: str, value: str, position: int | None = None, after: str | None = None):
        self.name = name
        self.value = value
        self.position = position
        self.after = after

    def description(self) -> str:
        if self.after:
            return f"Append {self.value} to {self.name} after {self.after}"
        pos = f" at position {self.position}" if self.position is not None else ""

        return f"Append {self.value} to {self.name}{pos}"

    def apply(self, content: str) -> str:
        """Add a value to a list in Python code"""

        tree = cst.parse_module(content)
        transformer = self.AddToListTransformer(self.name, self.value, self.position, self.after)
        modified_tree = tree.visit(transformer)

        if not transformer.found:
            raise ValueError(f"List '{self.name}' not found in file")

        return modified_tree.code


class RemoveFromList(PythonOperation):
    """Remove a value from a Python list"""

    class RemoveFromListTransformer(cst.CSTTransformer):
        """CST transformer to remove items from a list"""

        def __init__(self, list_name: str, value: str):
            self.list_name = list_name.split(".")
            self.value = value
            self.found = False
            self.removed = False
            self.current_name = []

        def visit_ClassDef(self, node: cst.ClassDef) -> bool | None:  # noqa: N802
            self.current_name.append(node.name.value)

            return True

        def leave_ClassDef(self, _: cst.ClassDef, updated_node: cst.CSTNode) -> cst.CSTNode:  # noqa: N802
            if self.current_name:
                self.current_name.pop()

            return updated_node

        def leave_Assign(self, _: cst.Assign, updated_node: cst.Assign) -> cst.CSTNode:  # noqa: N802
            # Skip if we're not at the right path depth
            if len(self.current_name) != len(self.list_name) - 1:
                return updated_node

            target = updated_node.targets[0].target

            # Handle direct name access
            if isinstance(target, cst.Name):
                target_name = target.value

                if self.current_name and self.current_name[0] == self.list_name[0]:
                    if target_name == self.list_name[-1] and isinstance(updated_node.value, cst.List):
                        self.found = True

                        return self._remove_from_list_node(updated_node)
                elif (
                    len(self.list_name) == 1
                    and target_name == self.list_name[0]
                    and isinstance(updated_node.value, cst.List)
                ):
                    self.found = True

                    return self._remove_from_list_node(updated_node)

            # Handle nested attributes
            elif isinstance(target, cst.Attribute):
                attr_parts = []
                node = target

                while isinstance(node, cst.Attribute):
                    attr_parts.append(node.attr.value)
                    node = node.value

                if isinstance(node, cst.Name):
                    attr_parts.append(node.value)
                    attr_parts.reverse()

                    if ".".join(attr_parts) == ".".join(self.list_name) and isinstance(updated_node.value, cst.List):
                        self.found = True

                        return self._remove_from_list_node(updated_node)

            return updated_node

        def _remove_from_list_node(self, node):
            """Remove the value from the list node"""

            # Parse the value to compare
            target_value = cst.parse_expression(self.value)

            # Filter out matching elements
            new_elements = []

            for element in node.value.elements:
                if not element.value.deep_equals(target_value):
                    new_elements.append(element)
                else:
                    self.removed = True

            return node.with_changes(value=node.value.with_changes(elements=new_elements))

    def __init__(self, list_name: str, value: str):
        self.list_name = list_name
        self.value = value

    def description(self) -> str:
        return f"Remove {self.value} from {self.list_name}"

    def apply(self, content: str) -> str:
        """Remove a value from a list in Python code"""

        tree = cst.parse_module(content)
        transformer = self.RemoveFromListTransformer(self.list_name, self.value)
        modified_tree = tree.visit(transformer)

        if not transformer.found:
            raise ValueError(f"List '{self.list_name}' not found in file")
        if not transformer.removed:
            raise ValueError(f"Value {self.value} not found in '{self.list_name}'")

        return modified_tree.code


class GetVariable(PythonOperation):
    """Get the value of a Python variable"""

    class GetVariableVisitor(cst.CSTVisitor):
        """CST visitor to extract variable values"""

        def __init__(self, name: str):
            self.name = name.split(".")
            self.value = None
            self.current_name = []

        def visit_ClassDef(self, node: cst.ClassDef) -> bool | None:  # noqa: N802
            self.current_name.append(node.name.value)
            return True

        def leave_ClassDef(self, node: cst.ClassDef) -> None:  # noqa: N802, ARG002
            if self.current_name:
                self.current_name.pop()

        def visit_Assign(self, node: cst.Assign) -> bool | None:  # noqa: N802
            # Skip if we're not at the right path depth
            if len(self.current_name) != len(self.name) - 1:
                return True

            target = node.targets[0].target

            # Handle direct name access (e.g., STATIC_ROOT)
            if isinstance(target, cst.Name):
                target_name = target.value

                # If we're in a class context, check if the class name matches
                if self.current_name and self.current_name[0] == self.name[0]:
                    if target_name == self.name[-1]:
                        self.value = cst.Module([]).code_for_node(node.value)
                        return False  # Stop visiting once found

                # If not in a class context, just match the name
                elif len(self.name) == 1 and target_name == self.name[0]:
                    self.value = cst.Module([]).code_for_node(node.value)
                    return False  # Stop visiting once found

            # Handle nested attributes (e.g., Settings.STATIC_ROOT)
            elif isinstance(target, cst.Attribute):
                attr_parts = []
                attr_node = target

                while isinstance(attr_node, cst.Attribute):
                    attr_parts.append(attr_node.attr.value)
                    attr_node = attr_node.value

                if isinstance(attr_node, cst.Name):
                    attr_parts.append(attr_node.value)
                    attr_parts.reverse()

                    # Check if the full path matches our target
                    if ".".join(attr_parts) == ".".join(self.name):
                        self.value = cst.Module([]).code_for_node(node.value)
                        return False  # Stop visiting once found

            return True

    def __init__(self, name: str):
        self.name = name

    def description(self) -> str:
        return f"Get value of {self.name}"

    def apply(self, content: str) -> str:
        """Get the value of a variable from Python code"""

        tree = cst.parse_module(content)
        visitor = self.GetVariableVisitor(self.name)
        tree.visit(visitor)

        if visitor.value is None:
            raise ValueError(f"Variable '{self.name}' not found in file")

        return visitor.value


class AssignVariable(PythonOperation):
    """Assign a value to a variable, creating it if it doesn't exist"""

    class AssignVariableTransformer(cst.CSTTransformer):
        def __init__(self, name: str, value: str):
            self.name = name
            self.value = value
            self.found = False

        def leave_Assign(self, original_node: cst.Assign, updated_node: cst.Assign) -> cst.CSTNode:  # noqa: N802, ARG002
            # Check if this assignment matches our target
            for target in updated_node.targets:
                if isinstance(target.target, cst.Name) and target.target.value == self.name:
                    self.found = True
                    # Replace value
                    value_node = cst.parse_expression(self.value)
                    return updated_node.with_changes(value=value_node)
            return updated_node

    def __init__(self, name: str, value: str | Any):
        self.name = name
        if not isinstance(value, str):
            value = repr(value)
        self.value = value

    def description(self) -> str:
        return f"Assign {self.value} to {self.name}"

    def apply(self, content: str) -> str:
        tree = cst.parse_module(content)
        transformer = self.AssignVariableTransformer(self.name, self.value)
        modified_tree = tree.visit(transformer)

        if not transformer.found:
            # Append to end of module
            assign_stmt = cst.parse_statement(f"{self.name} = {self.value}")
            # Add to body
            new_body = list(modified_tree.body)
            new_body.append(assign_stmt)
            modified_tree = modified_tree.with_changes(body=new_body)

        return modified_tree.code
